import * as d3 from "./helpers/d3-service"

import {comparators, keys} from "./helpers/constants"
import {ascendingComparator, descendingComparator, clamp, invertScale, sortData, getUnique} from "./helpers/common"
import cloneDeep from "lodash.clonedeep"


export default function DataGenerator () {
  /* eslint-disable no-magic-numbers */
  let config = {
    keyType: "number", // number, string, time,
    range: [0, 100],
    pointCount: 200,
    groupCount: 2,
    lineCount: 4,
    stringMinMaxLength: [4, 8],
    randomStepSize: 50,
    nullRatio: null
  }
  const cache = {
    data: null,
    baseDate: null
  }

  const DAY_IN_MS = 1000 * 60 * 60 * 24

  function generateRandomString (_length) {
    let stringLength = _length
    if (!_length) {
      const range = config.stringMinMaxLength
      stringLength = Math.round(Math.random() * (range[1] - range[0])) + range[0]
    }
    return [...Array(stringLength)].map(() => String.fromCharCode(Math.round(Math.random() * 25) + 97)).join("")
  }

  function generateSeries (_dataKeys, _range) {
    let value = d3.randomUniform(..._range)()
    const randomWalkStepSize = (_range[1] - _range[0]) / config.randomStepSize
    const rnd = d3.randomNormal(0, 1)
    return _dataKeys.map(d => {
      const isRandomNull = config.nullRatio && (Math.random() * 100 / config.nullRatio) < 1
      value = clamp(value + rnd() * randomWalkStepSize, _range)
      return {
        y: isRandomNull ? null : value,
        x: config.keyType === "time" ? d.toISOString() : d
      }
    })
  }

  function generateTestDataset () {
    let dataKeys = null
    if (config.keyType === "time") {
      cache.baseDate = new Date()
      const previousDate = new Date(cache.baseDate.getTime() - DAY_IN_MS * config.pointCount)
      dataKeys = d3.utcDay.range(previousDate, cache.baseDate)
    } else if (config.keyType === "string") {
      dataKeys = d3.range(0, config.pointCount).map(() => generateRandomString())
      dataKeys.sort((a, b) => a.localeCompare(b, "en", {numeric: false}))
    } else if (config.keyType === "number") {
      dataKeys = d3.range(0, config.pointCount).map((d, i) => i)
    }

    const series = d3.range(config.lineCount).map((d) => ({
      name: `Name ${d}`,
      label: `Label ${d}`,
      id: d,
      group: d < config.groupCount ? d : 0,
      values: generateSeries(dataKeys, config.range),
      measureName: `Measure ${d}`
    }))

    cache.data = {series}

    // console.log("generated data", cache.data)

    return cache.data
  }

  function setConfig (_config) {
    config = Object.assign({}, config, _config)
    return this
  }

  return {
    generateTestDataset,
    generateSeries,
    setConfig
  }
}

// accessors
const getKey = (d) => d[keys.KEY]
const getID = (d) => d[keys.ID]

export function augmentData (_data, _keyType, _sortBy, _fillData, _stackOffset, _yAxisPercentageFormat) {
  const dataBySeries = cloneDeep(_data[keys.SERIES])
  dataBySeries.forEach((serie) => {
    // convert type
    serie[keys.VALUES].forEach((d) => {
      if (_keyType === "time") {
        d[keys.KEY] = convertToDate(d[keys.KEY])
      }
      if (_fillData) {
        d[keys.VALUE] = Number(d[keys.VALUE])
      }
    })
  })
  const flatData = []

  // get all unique keys
  let allKeys = []
  dataBySeries.forEach(d => {
    allKeys = allKeys.concat(d.values)
  })
  allKeys = allKeys.map(d => d.key)
  allKeys = getUnique(allKeys, _keyType)

  // Normalize dataBySeries
  dataBySeries.forEach((serie) => {
    const keyValues = {}
    serie[keys.VALUES].forEach(d => {
      keyValues[d.key] = d.value
    })
    // fill data
    let filled = serie[keys.VALUES]
    if (_fillData) {
      filled = allKeys.map(d => ({
        key: d,
        value: (typeof keyValues[d] === "undefined") ? null : keyValues[d]
      }))
    }

    // sort
    serie[keys.VALUES] = sortData(filled, _keyType)
  })

  // flatten data
  dataBySeries.forEach((serie) => {
    serie[keys.VALUES].forEach((d) => {
      const dataPoint = {...serie}
      dataPoint[keys.KEY] = d[keys.KEY]
      dataPoint[keys.VALUE] = d[keys.VALUE]
      dataPoint[keys.COUNTVAL] = d[keys.COUNTVAL]
      dataPoint[keys.ABSOLUTEVAL] = d[keys.ABSOLUTEVAL]
      flatData.push(dataPoint)
    })
  })
  // sort flat data
  const flatDataSorted = sortData(flatData, _keyType)

  const dataByKey = d3.nest()
    .key(getKey)
    .entries(flatDataSorted.map(d => ({
      ...d,
      x: _keyType === "time" ? d.x.toISOString() : d.x
    })))
    .map((d) => {
      const dataPoint = {}
      dataPoint[keys.KEY] = _keyType === "time" ? new Date(d.key) : d.key
      dataPoint[keys.SERIES] = d.values
      return dataPoint
    })

  // get group keys
  const groupKeys = {}
  dataBySeries.forEach((d) => {
    if (!groupKeys[d[keys.GROUP]]) {
      groupKeys[d[keys.GROUP]] = []
    }
    groupKeys[d[keys.GROUP]].push(d[keys.ID])
  })

  // stack data
  const stackData = dataByKey
    .map((d) => {
      const points = {
        key: d[keys.KEY]
      }
      d.series.forEach((dB) => {
        points[dB[keys.ID]] = dB[keys.VALUE]
      })
      return points
    })

  // d3 stack
  const stack = d3.stack()
    .keys(dataBySeries.map(getID))
    .value((d, key) => d[key] || 0)
    .order(d3.stackOrderNone)
    .offset(d3[_stackOffset])

  // get stack totals
  const allKeyTotals = dataByKey.map(d => ({
    key: d[keys.KEY],
    total: d3.sum(d[keys.SERIES].map(dB => dB[typeof _yAxisPercentageFormat === "string" ? keys.ABSOLUTEVAL : keys.VALUE])),
    ...(d[keys.SERIES][0][keys.COUNTVAL] && {countval: d3.sum(d[keys.SERIES].map(dB => dB[keys.COUNTVAL]))})
  }))

  // sort
  switch (_sortBy) {
  case comparators.TOTAL_ASCENDING:
    allKeyTotals.sort(ascendingComparator("total", _keyType))
    break
  case comparators.TOTAL_DESCENDING:
    allKeyTotals.sort(descendingComparator("total", _keyType))
    break
  case comparators.ALPHA_ASCENDING:
    allKeyTotals.sort(ascendingComparator("key", _keyType))
    break
  case comparators.ALPHA_DESCENDING:
    allKeyTotals.sort(descendingComparator("key", _keyType))
    break
  case comparators.COUNTVAL_ASCENDING:
    allKeyTotals.sort(ascendingComparator("countval", _keyType))
    break
  case comparators.COUNTVAL_DESCENDING:
    allKeyTotals.sort(descendingComparator("countval", _keyType))
    break
  default:
    break
  }

  return {dataBySeries, dataByKey, stack, stackData, flatDataSorted, groupKeys, allKeyTotals}
}

export function getNearestDataPoint (_mouseX, _dataObject, _scales, _keyType) {
  const keyFromInvertedX = invertScale(_scales.xScale, _mouseX, _keyType)

  if (_keyType === "string") {
    // if we are keying on strings, simply find the value via a key match
    return _dataObject.dataByKey.find(d => d[keys.KEY] === keyFromInvertedX)
  }

  const bisectLeft = d3.bisector(getKey).left
  const dataEntryIndex = bisectLeft(_dataObject.dataByKey, keyFromInvertedX)
  const dataEntryForXPosition = _dataObject.dataByKey[dataEntryIndex]
  const dataEntryForXPositionPrev = _dataObject.dataByKey[Math.max(dataEntryIndex - 1, 0)]

  let nearestDataPoint = null
  if (keyFromInvertedX && dataEntryForXPosition && dataEntryForXPositionPrev) {
    if ((keyFromInvertedX - dataEntryForXPositionPrev[keys.KEY])
        < (dataEntryForXPosition[keys.KEY] - keyFromInvertedX)) {
      nearestDataPoint = dataEntryForXPositionPrev
    } else {
      nearestDataPoint = dataEntryForXPosition
    }
  }
  return nearestDataPoint
}

function convertToDate (_date) {
  // hacks to handle invalid date like "0014-06-08T00:00:00.000Z"
  return new Date(new Date(_date).toISOString())
}
